"""
Appropriate settings to run during development.

When running in development mode, selected.py should point to this file.
"""

from debusine.signing.settings import defaults

try:
    from debusine.signing.settings.db_postgresql import DATABASES
except (ModuleNotFoundError, PermissionError):
    from debusine.signing.settings.db_postgresql_default import DATABASES

__all__ = [
    'DATABASES',
    'DEBUG',
    'INSTALLED_APPS',
]

DEBUG = True

INSTALLED_APPS = defaults.INSTALLED_APPS.copy()
# INSTALLED_APPS.append('debug_toolbar')



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""debusine-admin command to edit workers metadata."""

import warnings
from typing import Any, NoReturn, override

from django.core.management import CommandParser

from debusine.server.management.commands import worker


class Command(worker.Command):
    """Command to edit the metadata of a Worker."""

    help = """
    Edit worker's metadata.

    Deprecated; use `debusine-admin worker edit_metadata` instead.
    """

    @override
    def add_arguments(self, parser: CommandParser) -> None:
        parser.add_argument(
            '--set',
            help=(
                "Filename with the metadata in YAML"
                " (note: its contents will be displayed to users in the web UI)"
            ),
            metavar='PATH',
        )
        parser.add_argument(
            'worker_name',
            help='Name of the worker of which the metadata will be edited',
        )

    @override
    def handle(self, *args: Any, **options: Any) -> NoReturn:
        """Forward to `debusine-admin worker edit_metadata`."""
        warnings.warn(
            "The `debusine-admin edit_worker_metadata` command has been "
            "deprecated in favour of `debusine-admin worker edit_metadata`",
            DeprecationWarning,
        )
        options = dict(options)
        options["worker"] = options["worker_name"]
        del options["worker_name"]
        super().handle_edit_metadata(*args, **options)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for Debusine CLI suite management commands."""

from typing import Any
from unittest import mock

import yaml

from debusine.artifacts.models import CollectionCategory
from debusine.client.commands.archives.suites import Create
from debusine.client.commands.tests.base import BaseCliTests
from debusine.client.debusine import Debusine
from debusine.client.exceptions import DebusineError
from debusine.client.models import (
    CollectionData,
    CollectionDataNew,
    RuntimeParameter,
    WorkflowTemplateData,
    WorkflowTemplateDataNew,
    WorkspaceInheritanceChain,
    WorkspaceInheritanceChainElement,
)


class CreateTests(BaseCliTests):
    """Tests for ``debusine archive suite create``."""

    def cmdline(self, *args: str) -> list[str]:
        """Shortcut to build a command."""
        return ["archive", "suite", "create", *args]

    def chain(
        self,
        *args: WorkspaceInheritanceChainElement,
    ) -> WorkspaceInheritanceChain:
        """Shortcut to build an inheritance chain."""
        return WorkspaceInheritanceChain(chain=list(args))

    def suite(self, name: str, data: dict[str, Any]) -> CollectionData:
        """Shortcut to build a suite collection."""
        return CollectionData(
            id=1,
            name=name,
            category=CollectionCategory.SUITE,
            data=data,
            links={"webui_self": "/c/1"},
        )

    def workflow_template(
        self,
        name: str,
        task_name: str,
        workflow_template_id: int = 1,
        static_parameters: dict[str, Any] | None = None,
    ) -> WorkflowTemplateData:
        """Shortcut to build a workflow template."""
        return WorkflowTemplateData(
            id=workflow_template_id,
            name=name,
            task_name=task_name,
            static_parameters=static_parameters or {},
            runtime_parameters=RuntimeParameter.ANY,
            links={"webui_self": f"/wt/{workflow_template_id}"},
        )

    def test_base_workflow_template_nonexistent_workspace(self) -> None:
        command = self.create_command(
            self.cmdline(
                "--workspace=nonexistent",
                "--architecture=amd64",
                "--base-workflow-template=upload-to-unstable",
                "sid-overlay",
            )
        )
        assert isinstance(command, Create)

        with (
            mock.patch.object(
                Debusine,
                "workflow_template_get",
                side_effect=DebusineError(title="Workspace not found"),
            ),
            mock.patch.object(
                Debusine,
                "get_workspace_inheritance",
                side_effect=DebusineError(title="Workspace not found"),
            ),
            self.assertRaises(DebusineError) as raised,
        ):
            self.capture_output(
                command.get_base_workflow_template, assert_system_exit_code=3
            )

        self.assertDebusineError(
            raised.exception, {"title": "Workspace not found"}
        )

    def test_base_workflow_template_in_different_scope(self) -> None:
        command = self.create_command(
            self.cmdline(
                "--workspace=workspace",
                "--architecture=amd64",
                "--base-workflow-template=upload-to-unstable",
                "sid-overlay",
            )
        )
        assert isinstance(command, Create)
        El = WorkspaceInheritanceChainElement
        chain = self.chain(El(scope="other-scope", workspace="base"))

        with (
            mock.patch.object(
                Debusine,
                "workflow_template_get",
                side_effect=DebusineError(title="Object not found"),
            ),
            mock.patch.object(
                Debusine, "get_workspace_inheritance", return_value=chain
            ),
        ):
            stderr, _ = self.capture_output(
                command.get_base_workflow_template, assert_system_exit_code=3
            )

        self.assertEqual(
            stderr,
            "Workflow template 'upload-to-unstable' not found in any parent "
            "workspace in this scope.\n",
        )

    def test_base_workflow_template_wrong_task_name(self) -> None:
        command = self.create_command(
            self.cmdline(
                "--workspace=workspace",
                "--architecture=amd64",
                "--base-workflow-template=sample",
                "sid-overlay",
            )
        )
        assert isinstance(command, Create)

        with mock.patch.object(
            Debusine,
            "workflow_template_get",
            return_value=self.workflow_template("sample", "noop"),
        ):
            stderr, _ = self.capture_output(
                command.get_base_workflow_template, assert_system_exit_code=3
            )

        self.assertEqual(
            stderr,
            "Workflow template 'sample' is not a debian_pipeline template.\n",
        )

    def test_base_workflow_template_in_current_workspace(self) -> None:
        command = self.create_command(
            self.cmdline(
                "--architecture=amd64",
                "--base-workflow-template=upload-to-unstable",
                "sid-overlay",
            )
        )
        assert isinstance(command, Create)
        base_workflow_template = self.workflow_template(
            "upload-to-unstable", "debian_pipeline"
        )

        with mock.patch.object(
            Debusine,
            "workflow_template_get",
            return_value=base_workflow_template,
        ) as workflow_template_get:
            self.assertEqual(
                command.get_base_workflow_template(), base_workflow_template
            )

        workflow_template_get.assert_called_once_with(
            "developers", "upload-to-unstable"
        )

    def test_base_workflow_template_in_parent_workspace(self) -> None:
        command = self.create_command(
            self.cmdline(
                "--workspace=workspace",
                "--architecture=amd64",
                "--base-workflow-template=upload-to-unstable",
                "sid-overlay",
            )
        )
        assert isinstance(command, Create)
        El = WorkspaceInheritanceChainElement
        chain = self.chain(El(scope=self.default_server, workspace="base"))
        base_workflow_template = self.workflow_template(
            "upload-to-unstable", "debian_pipeline"
        )

        with (
            mock.patch.object(
                Debusine,
                "workflow_template_get",
                side_effect=iter(
                    [
                        DebusineError(title="Object not found"),
                        base_workflow_template,
                    ]
                ),
            ) as workflow_template_get,
            mock.patch.object(
                Debusine, "get_workspace_inheritance", return_value=chain
            ),
        ):
            self.assertEqual(
                command.get_base_workflow_template(), base_workflow_template
            )

        self.assertEqual(workflow_template_get.call_count, 2)
        workflow_template_get.assert_has_calls(
            [
                mock.call("workspace", "upload-to-unstable"),
                mock.call("base", "upload-to-unstable"),
            ]
        )

    def test_adjust_static_parameters_with_architecture_all(self) -> None:
        command = self.create_command(
            self.cmdline(
                "--workspace=workspace",
                "--architecture=all",
                "--architecture=amd64",
                "--base-workflow-template=upload-to-unstable",
                "sid-overlay",
            )
        )
        assert isinstance(command, Create)

        self.assertEqual(
            command.adjust_static_parameters(
                {"vendor": "debian", "enable_upload": True}
            ),
            {
                "vendor": "debian",
                "suite": f"sid-overlay@{CollectionCategory.SUITE}",
                "architectures_allowlist": ["all", "amd64"],
            },
        )

    def test_adjust_static_parameters_without_architecture_all(self) -> None:
        command = self.create_command(
            self.cmdline(
                "--workspace=workspace",
                "--architecture=amd64",
                "--architecture=i386",
                "--base-workflow-template=upload-to-unstable",
                "sid-overlay",
            )
        )
        assert isinstance(command, Create)

        self.assertEqual(
            command.adjust_static_parameters(
                {"vendor": "debian", "enable_upload": True}
            ),
            {
                "vendor": "debian",
                "suite": f"sid-overlay@{CollectionCategory.SUITE}",
                "architectures_allowlist": ["all", "amd64", "i386"],
            },
        )

    def test_run_create_collection_error(self) -> None:
        cli = self.create_cli(
            self.cmdline(
                "--workspace=workspace", "--architecture=amd64", "sid-overlay"
            )
        )

        with mock.patch.object(
            Debusine,
            "collection_create",
            side_effect=DebusineError(title="expected error"),
        ):
            exception = self.assertShowsError(cli.execute)

        self.assertDebusineError(exception, {"title": "expected error"})

    def test_run_create_workflow_template_error(self) -> None:
        cli = self.create_cli(
            self.cmdline(
                "--workspace=workspace",
                "--architecture=amd64",
                "--base-workflow-template=upload-to-unstable",
                "sid-overlay",
            )
        )
        base_workflow_template = self.workflow_template(
            "upload-to-unstable", "debian_pipeline"
        )
        suite = self.suite(
            "sid-overlay", {"components": "main", "architectures": ["amd64"]}
        )

        with (
            mock.patch.object(
                Debusine,
                "workflow_template_get",
                return_value=base_workflow_template,
            ),
            mock.patch.object(
                Debusine, "collection_create", return_value=suite
            ),
            mock.patch.object(
                Debusine,
                "workflow_template_create",
                side_effect=DebusineError(title="expected error"),
            ),
        ):
            exception = self.assertShowsError(cli.execute)

        self.assertDebusineError(exception, {"title": "expected error"})

    def test_run_without_base_workflow_template(self) -> None:
        cli = self.create_cli(
            self.cmdline(
                "--workspace=workspace",
                "--architecture=amd64",
                "--architecture=i386",
                "sid-overlay",
            )
        )
        expected_suite_data = {
            "components": ["main"],
            "architectures": ["amd64", "i386"],
        }
        suite = self.suite("sid-overlay", expected_suite_data)

        with mock.patch.object(
            Debusine, "collection_create", return_value=suite
        ) as collection_create:
            stderr, stdout = self.capture_output(cli.execute)

        collection_create.assert_called_once_with(
            "workspace",
            CollectionDataNew(
                name="sid-overlay",
                category=CollectionCategory.SUITE,
                data=expected_suite_data,
            ),
        )
        self.assertEqual(stderr, "")
        self.assertEqual(
            yaml.safe_load(stdout),
            {"result": "success", "suite": "https://debusine.debian.org/c/1"},
        )

    def test_run_with_base_workflow_template(self) -> None:
        cli = self.create_cli(
            self.cmdline(
                "--workspace=workspace",
                "--architecture=amd64",
                "--architecture=i386",
                "--base-workflow-template=upload-to-unstable",
                "sid-overlay",
            )
        )
        base_workflow_template = self.workflow_template(
            "upload-to-unstable",
            "debian_pipeline",
            static_parameters={"vendor": "debian", "enable_upload": True},
        )
        expected_suite_data = {
            "components": ["main"],
            "architectures": ["amd64", "i386"],
        }
        suite = self.suite("sid-overlay", expected_suite_data)
        expected_static_parameters = {
            "vendor": "debian",
            "suite": f"sid-overlay@{CollectionCategory.SUITE}",
            "architectures_allowlist": ["all", "amd64", "i386"],
        }
        workflow_template = self.workflow_template(
            "publish-to-sid-overlay",
            "debian_pipeline",
            workflow_template_id=2,
            static_parameters=expected_static_parameters,
        )

        with (
            mock.patch.object(
                Debusine,
                "workflow_template_get",
                return_value=base_workflow_template,
            ),
            mock.patch.object(
                Debusine, "collection_create", return_value=suite
            ) as collection_create,
            mock.patch.object(
                Debusine,
                "workflow_template_create",
                return_value=workflow_template,
            ) as workflow_template_create,
        ):
            stderr, stdout = self.capture_output(cli.execute)

        collection_create.assert_called_once_with(
            "workspace",
            CollectionDataNew(
                name="sid-overlay",
                category=CollectionCategory.SUITE,
                data=expected_suite_data,
            ),
        )
        workflow_template_create.assert_called_once_with(
            "workspace",
            WorkflowTemplateDataNew(
                name="publish-to-sid-overlay",
                task_name="debian_pipeline",
                static_parameters=expected_static_parameters,
                runtime_parameters=RuntimeParameter.ANY,
                priority=0,
            ),
        )
        self.assertEqual(stderr, "")
        self.assertEqual(
            yaml.safe_load(stdout),
            {
                "result": "success",
                "suite": "https://debusine.debian.org/c/1",
                "workflow_template": "https://debusine.debian.org/wt/2",
            },
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Views for the server application: workspaces."""

import logging

from rest_framework import status
from rest_framework.parsers import JSONParser
from rest_framework.request import Request
from rest_framework.response import Response

from debusine.client.models import (
    WorkspaceInheritanceChain,
    WorkspaceInheritanceChainElement,
)
from debusine.db.context import context
from debusine.db.models.workspaces import Workspace, WorkspaceChain
from debusine.server.exceptions import DebusineAPIException
from debusine.server.serializers import WorkspaceChainSerializer
from debusine.server.views.base import BaseAPIView
from debusine.server.views.rest import IsUserTokenAuthenticated

logger = logging.getLogger(__name__)


class WorkspaceInheritanceView(BaseAPIView):
    """View used to fetch and update a workspace inheritance chain."""

    permission_classes = [IsUserTokenAuthenticated]
    parser_classes = [JSONParser]

    def _get_chain(self, workspace: Workspace) -> WorkspaceInheritanceChain:
        res = WorkspaceInheritanceChain(chain=[])
        for pk, scope_name, workspace_name in (
            WorkspaceChain.objects.filter(child=workspace)
            .order_by("order")
            .values_list("parent__id", "parent__scope__name", "parent__name")
        ):
            res.chain.append(
                WorkspaceInheritanceChainElement(
                    id=pk, scope=scope_name, workspace=workspace_name
                )
            )
        return res

    def get(
        self,
        request: Request,  # noqa: ARG002, U100
        workspace: str,
    ) -> Response:
        """Retrieve the contents of the collection."""
        self.set_current_workspace(workspace)
        assert context.workspace
        self.enforce(context.workspace.can_display)
        chain = self._get_chain(context.require_workspace())
        return Response(
            chain.model_dump(mode="json"), status=status.HTTP_200_OK
        )

    def post(self, request: Request, workspace: str) -> Response:
        """Update the contents of the collection."""
        self.set_current_workspace(workspace)
        current = context.require_workspace()
        self.enforce(current.can_configure)

        # Decode request.data
        serializer = WorkspaceChainSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        user_chain = serializer.validated_data["chain"]

        # Check permissions of items
        existing_ids = frozenset(c.id for c in self._get_chain(current).chain)
        for ws in user_chain:
            self.enforce(ws.can_display)
            if (
                ws.pk not in existing_ids
                and not ws.public
                and ws.scope != current.scope
            ):
                raise DebusineAPIException(
                    title="Workspace cannot be inherited",
                    detail=(
                        "Private workspaces from other scopes"
                        " cannot be inherited."
                    ),
                    status_code=status.HTTP_403_FORBIDDEN,
                )

        # Set new chain
        try:
            current.set_inheritance(user_chain)
        except ValueError as e:
            raise DebusineAPIException(
                title="Invalid inheritance chain",
                detail=str(e),
                status_code=status.HTTP_400_BAD_REQUEST,
            )

        # Return the new chain
        chain = self._get_chain(current)
        return Response(
            chain.model_dump(mode="json"), status=status.HTTP_200_OK
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Debusine command line interface, task configuration management commands."""

import argparse
from pathlib import Path
from typing import override

from debusine.client.commands.base import DebusineCommand, WorkspaceCommand
from debusine.client.task_configuration import (
    InvalidRepository,
    LocalTaskConfigurationRepository,
)


class Pull(WorkspaceCommand, group="task-config"):
    """
    Check out a task configuration collection.

    Update a local checkout if it exists.
    """

    local_repo: LocalTaskConfigurationRepository
    collection: str

    @override
    @classmethod
    def configure(cls, parser: argparse.ArgumentParser) -> None:
        super().configure(parser)
        parser.add_argument(
            "collection",
            type=str,
            nargs="?",
            help="collection name (not needed to refresh an existing checkout)",
        )
        parser.add_argument(
            "--workdir",
            type=Path,
            default=Path.cwd(),
            help="working directory to use. Default: current directory.",
        )

    def __init__(
        self, parser: argparse.ArgumentParser, args: argparse.Namespace
    ) -> None:
        """Initialize a local repository."""
        super().__init__(parser, args)

        if not LocalTaskConfigurationRepository.is_checkout(self.args.workdir):
            self.args.workdir.mkdir(exist_ok=True)
            self.local_repo = LocalTaskConfigurationRepository.from_path(
                self.args.workdir
            )
            self.collection = self.args.collection or "default"
        else:
            self.local_repo = LocalTaskConfigurationRepository.from_path(
                self.args.workdir
            )
            manifest = self.local_repo.manifest
            assert manifest is not None
            if self.workspace_arg_set:
                if self.workspace != manifest.workspace:
                    self._fail(
                        f"--workspace={self.workspace} is provided,"
                        " but the repository checked out refers to"
                        f" workspace {manifest.workspace!r}"
                    )
            else:
                self.workspace = manifest.workspace

            if (
                self.args.collection is not None
                and self.args.collection != manifest.collection.name
            ):
                self._fail(
                    f"--collection={self.args.collection} is provided,"
                    " but the repository checked out refers to"
                    f" collection {manifest.collection.name!r}"
                )
            self.collection = manifest.collection.name

    @override
    def run(self) -> None:
        server_repo = self.debusine.fetch_task_configuration_collection(
            workspace=self.workspace, name=self.collection
        )
        try:
            stats = self.local_repo.pull(
                server_repo, logger=self.debusine._logger
            )
        except InvalidRepository as e:
            self._fail(str(e))
        self.debusine._logger.info(
            "%d added, %d updated, %d deleted, %d unchanged",
            stats.added,
            stats.updated,
            stats.deleted,
            stats.unchanged,
        )


class LegacyPull(
    Pull, name="task-config-pull", deprecated="see `task-config pull`"
):
    """
    Check out a task configuration collection.

    Update a local checkout if it exists.
    """


class Push(DebusineCommand, group="task-config"):
    """
    Send local task configuration data to Debusine.

    Replace the contents of the collection on the server with those of the
    local checkout.
    """

    @override
    @classmethod
    def configure(cls, parser: argparse.ArgumentParser) -> None:
        super().configure(parser)
        parser.add_argument(
            "--workdir",
            type=Path,
            default=Path.cwd(),
            help="working directory to use. Default: current directory.",
        )
        parser.add_argument(
            "-n",
            "--dry-run",
            action="store_true",
            help="do not update the collection:"
            " only show what would be updated",
        )
        parser.add_argument(
            "-f",
            "--force",
            action="store_true",
            help="push the collection, skipping all checks",
        )

    @override
    def run(self) -> None:
        workdir = self.args.workdir
        dry_run = self.args.dry_run
        force = self.args.force
        if not LocalTaskConfigurationRepository.is_checkout(workdir):
            self._fail(f"{workdir} is not a repository")
        local_repo = LocalTaskConfigurationRepository.from_path(workdir)
        assert local_repo.manifest is not None

        if local_repo.is_dirty():
            message = (
                f"{workdir} has uncommitted changes:"
                " please commit them before pushing"
            )
            if force:
                self.debusine._logger.warning("%s", message)
            else:
                self._fail(message)

        server_repo = self.debusine.fetch_task_configuration_collection(
            workspace=local_repo.manifest.workspace,
            name=local_repo.manifest.collection.name,
        )

        server_git_commit = server_repo.manifest.collection.data.get(
            "git_commit"
        )
        if server_git_commit is not None and not local_repo.has_commit(
            server_git_commit
        ):
            message = (
                "server collection was pushed from commit"
                f" {server_git_commit} which is not known to {workdir}"
            )
            if force:
                self.debusine._logger.warning("%s", message)
            else:
                self._fail(message)

        if (new_git_commit := local_repo.git_commit()) is not None:
            local_repo.manifest.collection.data["git_commit"] = new_git_commit
        else:
            local_repo.manifest.collection.data.pop("git_commit", None)

        if dry_run:
            self.debusine._logger.info("Pushing data to server (dry run)...")
        else:
            self.debusine._logger.info("Pushing data to server...")

        results = self.debusine.push_task_configuration_collection(
            repo=local_repo, dry_run=dry_run
        )
        self.debusine._logger.info(
            "%d added, %d updated, %d removed, %d unchanged",
            results.added,
            results.updated,
            results.removed,
            results.unchanged,
        )


class LegacyPush(
    Push, name="task-config-push", deprecated="see `task-config push`"
):
    """
    Send local task configuration data to Debusine.

    Replace the contents of the collection on the server with those of the
    local checkout.
    """



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the management command list_workers."""

from debusine.django.management.tests import call_command
from debusine.server.management.commands.tests.utils import TabularOutputTests
from debusine.test.django import TestCase


class ListWorkersCommandTests(TabularOutputTests, TestCase):
    """Test for list_workers management command."""

    def test_forwards_with_deprecation_warning(self) -> None:
        """``list_workers`` forwards to ``worker list`` and warns."""
        worker_1 = self.playground.create_worker(fqdn="recent-ping")
        assert worker_1.token is not None
        worker_1.mark_connected()
        assert worker_1.connected_at is not None

        with (
            self.assertWarnsMessage(
                DeprecationWarning,
                "The `debusine-admin list_workers` command has been "
                "deprecated in favour of `debusine-admin worker list`",
            ),
            self.assertPrintsTable() as output,
        ):
            call_command("list_workers")
        self.assertEqual(output.col(0), [worker_1.name])
        self.assertEqual(output.col(2), [worker_1.registered_at.isoformat()])
        self.assertEqual(output.col(3), [worker_1.connected_at.isoformat()])
        self.assertEqual(output.col(4), [worker_1.token.hash])
        self.assertEqual(output.col(5), [str(worker_1.token.enabled)])



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
Server-side database interaction for worker tasks.

Worker tasks mostly run on external workers, but they also run some
pre-dispatch code on the server which needs to access the database.  To
avoid having to import Django from :py:mod:`debusine.tasks`, we define an
interface that is implemented by server code.
"""

from abc import ABC, abstractmethod
from collections.abc import Collection as AbcCollection
from typing import Any, TYPE_CHECKING, overload

from debusine.artifacts.models import ArtifactCategory, CollectionCategory
from debusine.client.models import RelationType
from debusine.tasks.models import (
    BackendType,
    ExtraExternalRepository,
    InputArtifactMultiple,
    InputArtifactSingle,
    InputCollectionSingle,
    LookupMultiple,
    LookupSingle,
)

if TYPE_CHECKING:
    from debusine.tasks.executors import ExecutorImageCategory
    from debusine.tasks.inputs import inputs


class FieldResolver(ABC):
    """Interface for interacting with the database from worker tasks."""

    @overload
    @abstractmethod
    def lookup_single_artifact(
        self,
        lookup: LookupSingle,
        *,
        default_category: CollectionCategory | None = None,
        label: str,
    ) -> InputArtifactSingle: ...

    @overload
    @abstractmethod
    def lookup_single_artifact(
        self,
        lookup: None,
        *,
        default_category: CollectionCategory | None = None,
        label: str,
    ) -> None: ...

    @abstractmethod
    def lookup_single_artifact(
        self,
        lookup: LookupSingle | None,
        *,
        default_category: CollectionCategory | None = None,
        label: str,
    ) -> InputArtifactSingle | None:
        """
        Look up a single artifact.

        :param lookup: A :ref:`lookup-single`.
        :param default_category: If the first segment of a string lookup
          (which normally identifies a collection) does not specify a
          category, use this as the default category.
        :return: Information about the artifact, or None if the provided
          lookup is None (for convenience in some call sites).
        :raises KeyError: if the lookup does not resolve to an item.
        :raises LookupError: if the lookup is invalid in some way, or does
          not resolve to an artifact.
        """

    @abstractmethod
    def lookup_multiple_artifacts(
        self,
        lookup: LookupMultiple,
        *,
        default_category: CollectionCategory | None = None,
        label: str,
    ) -> InputArtifactMultiple:
        """
        Look up multiple artifacts.

        :param lookup: A :ref:`lookup-multiple`.
        :param default_category: If the first segment of a string lookup
          (which normally identifies a collection) does not specify a
          category, use this as the default category.
        :return: Information about each artifact.
        :raises KeyError: if any of the lookups does not resolve to an item.
        :raises LookupError: if any of the lookups is invalid in some way,
          or does not resolve to an artifact.
        """

    @abstractmethod
    def find_related_artifacts(
        self,
        artifact_ids: AbcCollection[int],
        *,
        target_category: ArtifactCategory,
        relation_type: RelationType = RelationType.RELATES_TO,
        label: str,
    ) -> InputArtifactMultiple:
        """
        Find artifacts via relations.

        For each of the artifacts in ``artifact_ids``, find all artifacts of
        ``target_category`` that are related to the initial artifacts with a
        relation of type ``relation_type``.

        :param artifact_ids: The IDs of artifacts to start from.
        :param target_category: The category of artifacts to find.
        :param relation_type: The type of relation to follow.
        :return: Information about each artifact.
        """

    @abstractmethod
    def get_server_setting(self, setting: str) -> str:
        """Look up a Django setting (strings only)."""

    @abstractmethod
    def get_fallback_architecture(self) -> str:
        """Look up the fallback architecture."""

    @abstractmethod
    def export_suite_signing_keys(self, suite_id: int) -> str | None:
        """Export signing keys for a suite as an ASCII-armored keyring."""

    def resolve_inputs(
        self,
        container: "inputs.TaskFieldContainer",
        stage: "inputs.Stage | None" = None,
    ) -> None:
        """
        Resolve inputs for this task.

        This will look up information required by the input using this task
        database, and make it available to task methods.
        """
        for cont, inp in container.iter_all_fields(stage):
            try:
                cont.__dict__[inp.name] = inp.resolve(self)
            except Exception as exc:
                exc.add_note(f"Task input field: {inp.name}")
                raise exc

    @abstractmethod
    def lookup_environment(
        self,
        lookup: LookupSingle,
        *,
        architecture: str | None = None,
        backend: BackendType | None = None,
        default_category: CollectionCategory | None = None,
        image_category: "ExecutorImageCategory | None" = None,
        try_variant: str | None = None,
        label: str,
    ) -> InputArtifactSingle:
        """
        Lookup an environment.

        This automatically fills in some additional constraints if needed.

        :param lookup: the base lookup provided by the task data
        :param architecture: the task's build architecture, if available
        :param backend: the task's backend, or None if the environment lookup
          does not need to be constrained to a particular backend
        :param default_category: the default category to use for the first
          segment of the lookup
        :param image_category: try to use an environment with this image
          category; defaults to the image category needed by the executor for
          `self.backend`
        :param try_variant: None if the environment lookup does not need to be
          constrained to a particular variant; otherwise, and if `lookup` does
          not already specify a variant, then try looking up an environment with
          this variant first, and fall back to looking for an environment with
          no variant
        :return: the InputArtifactSingle of a suitable environment artifact
        """

    @abstractmethod
    def resolve_environmentinput(
        self, field: "inputs.EnvironmentInput"
    ) -> InputArtifactSingle:
        """Resolve the value for an EnvironmentInput field."""

    @abstractmethod
    def resolve_optionalenvironmentinput(
        self, field: "inputs.OptionalEnvironmentInput"
    ) -> InputArtifactSingle | None:
        """Resolve the value for an OptionalEnvironmentInput field."""

    @abstractmethod
    def resolve_suitearchiveinput(
        self, field: "inputs.SuiteArchiveInput"
    ) -> tuple[InputCollectionSingle, InputCollectionSingle | None]:
        """Resolve the value for an SuiteArchiveInput field."""

    @abstractmethod
    def resolve_extrarepositoriesinput(
        self, field: "inputs.ExtraRepositoriesInput"
    ) -> list[ExtraExternalRepository] | None:
        """Resolve the value for an ExtraRepositoriesInput field."""

    @abstractmethod
    def get_field[T](self, field: "inputs.DataFieldInput[T, Any]") -> T | None:
        """Lookup the given attribute in the input data."""

    def get_field_collection[T](
        self, field: "inputs.DataListFieldInput[T, Any]"
    ) -> AbcCollection[T]:
        """
        Lookup the given attribute in the input data.

        The value, if not None, is expected to be, and validated as, a
        collection of the given ``item_type``.
        """
        # Prevent import loop
        from debusine.tasks import TaskConfigError

        # get_field only uses isinstance, and mypy is being stricter here than
        # what we need, until there is a way of distinguishing types for
        # checking from types for instantiating
        values = self.get_field(field)
        if values is None:
            return []
        for idx, value in enumerate(values):
            if not isinstance(value, field.item_type):
                raise TaskConfigError(
                    f"Task data field {field.field}[{idx}] is {value!r}"
                    f" which is not of type {field.item_type}"
                )
        return values



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Typing utilities."""

from collections.abc import Callable
from typing import Any, ParamSpec, TypeVar, cast


# See https://stackoverflow.com/questions/71253495/how-to-annotate-the-type-of-arguments-forwarded-to-another-function  # noqa: E501
def copy_signature_from[T, **P](
    _origin: Callable[P, T],
) -> Callable[[Callable[..., Any]], Callable[P, T]]:
    """Copy function signature from another one."""

    def decorator(target: Callable[..., Any]) -> Callable[P, T]:
        return cast(Callable[P, T], target)

    return decorator


# This return type isn't very accurate, but this function is only used at
# run-time and not when type-checking anyway, so it doesn't matter much.
def patch_runtime_generic_support[T: type](real_class: T) -> type:
    """
    Patch run-time support for generic types into a class.

    Several libraries have type stubs that define generic classes even
    though their run-time equivalents do not support the necessary special
    method.  Code using these classes can work around this as follows::

        if TYPE_CHECKING:
            SomeClassBase: TypeAlias = SomeClass
        else:
            # Some classes don't support generic types at run-time yet.
            SomeClassBase = patch_runtime_generic_support(SomeClass)
    """

    class PatchRuntimeGenericSupport:
        def __class_getitem__(
            cls, *args: TypeVar | ParamSpec | tuple[TypeVar | ParamSpec, ...]
        ) -> type[T]:
            return real_class

    return PatchRuntimeGenericSupport



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Common infrastructure for server view tests."""

import logging
from collections.abc import Callable
from functools import partial
from typing import Any, overload, override
from unittest import mock

from django.test import Client
from pydantic import BaseModel
from requests.structures import CaseInsensitiveDict
from rest_framework import status

from debusine.client import exceptions
from debusine.client.debusine import Debusine
from debusine.client.debusine_http_client import DebusineHttpClient
from debusine.db.playground import scenarios
from debusine.test.django import TestCase as BaseTestCase
from debusine.test.django import TestResponseType
from debusine.utils.urls import DEFAULT_REQUEST_TIMEOUT


class DjangoDebusineHttpClient(DebusineHttpClient):
    """DebusineHttpClient using the Django test client."""

    def __init__(
        self,
        client: Client,
        api_url: str,
        token: str | None,
        scope: str | None = None,
        timeout: int | tuple[int, int] = DEFAULT_REQUEST_TIMEOUT,
    ) -> None:
        """Initialize DebusineHttpClient."""
        super().__init__(api_url, token, scope, timeout=timeout)
        self.client = client

    def _django_method(self, method: str) -> Callable[..., TestResponseType]:
        method_to_func: dict[str, Callable[..., TestResponseType]] = {
            'GET': self.client.get,
            'POST': self.client.post,
            'PUT': self.client.put,
            'PATCH': self.client.patch,
            'DELETE': self.client.delete,
        }
        return method_to_func[method]

    @overload
    def _api_request(
        self,
        method: str,
        url: str,
        expected_class: None,
        data: dict[str, Any] | None = None,
        *,
        expected_statuses: list[int] | None = None,
        if_match: str | None = None,
    ) -> tuple[None, CaseInsensitiveDict[str]]: ...

    @overload
    def _api_request[ClientModel: BaseModel](
        self,
        method: str,
        url: str,
        expected_class: type[ClientModel],
        data: dict[str, Any] | None = None,
        *,
        expected_statuses: list[int] | None = None,
        if_match: str | None = None,
    ) -> tuple[ClientModel, CaseInsensitiveDict[str]]: ...

    @override
    def _api_request[ClientModel: BaseModel](
        self,
        method: str,
        url: str,
        expected_class: type[ClientModel] | None,
        data: dict[str, Any] | None = None,
        *,
        if_match: str | None = None,
        expected_statuses: list[int] | None = None,
    ) -> tuple[ClientModel | None, CaseInsensitiveDict[str]]:
        """
        Request to the server.

        :param method: HTTP method (GET, POST, ...).
        :param url: The complete URL to request.
        :param expected_class: expected object class that the server.
          will return. Used to deserialize the response and return an object.
          If None expects an empty response body.
        :param if_match: if not None, set it to the header "If-Match"
        :param expected_statuses: defaults to expect HTTP 200. List of HTTP
          status codes that might be return by the call. If it receives
          an unexpected one it raises UnexpectedResponseError.
        :raises exceptions.UnexpectedResponseError: the server didn't return
          a valid JSON or returned an unexpected HTTP status code.
        :raises exceptions.WorkRequestNotFound: the server could not find the
          work request.
        :raises exceptions.ClientConnectionError: the client could not connect
          to the server.
        :raises ValueError: invalid options passed.
        :raises DebusineError: the server returned an error with
          application/problem+json. Contains the detail message.
        :raises exceptions.ClientForbiddenError: the server returned HTTP 403.
        :return: an object of the expected_class or None if expected_class
          was None.
        """
        # FIXME: there is extensive use of pragma: no cover to keep this
        # aligned with the base implementation. As more Debusine methods are
        # tested with thiswill, most of the pragmas will probably be dropped
        if data is not None and method == 'GET':  # pragma: no cover
            raise ValueError('data argument not allowed with HTTP GET')

        if expected_statuses is None:  # pragma: no cover
            expected_statuses = [status.HTTP_200_OK]

        optional_kwargs = {}

        if data is not None:
            optional_kwargs = {"data": data, "content_type": "application/json"}

        try:
            headers = {}
            if self._scope is not None:  # pragma: no cover
                headers["X-Debusine-Scope"] = self._scope
            if self._token is not None:  # pragma: no cover
                headers["Token"] = self._token
            if if_match is not None:
                headers["If-Match"] = if_match
            response = self._django_method(method)(
                url, headers=headers, **optional_kwargs
            )
        except ConnectionError as exc:  # pragma: no cover
            raise exceptions.ClientConnectionError(
                f'Cannot connect to {url}. Error: {exc}'
            ) from exc

        if error := self._django_debusine_problem(response):
            raise exceptions.DebusineError(**error)
        elif response.status_code in expected_statuses:
            return (
                self._handle_response_content(
                    response.content, expected_class, url
                ),
                CaseInsensitiveDict[str](response.headers),
            )
        elif (
            response.status_code == status.HTTP_404_NOT_FOUND
        ):  # pragma: no cover
            raise exceptions.NotFoundError(f'Not found ({url})')
        elif (
            response.status_code == status.HTTP_403_FORBIDDEN
        ):  # pragma: no cover
            raise exceptions.ClientForbiddenError(
                f"HTTP 403. Token ({self._token}) is invalid or disabled"
            )
        else:  # pragma: no cover
            raise exceptions.UnexpectedResponseError(
                f'Server returned unexpected status '
                f'code: {response.status_code} ({url})'
            )

    @staticmethod
    def _django_debusine_problem(
        response: TestResponseType,
    ) -> dict[Any, Any] | None:
        """If response is an application/problem+json returns the body JSON."""
        if response.headers.get("content-type") == "application/problem+json":
            content = response.json()
            assert isinstance(content, dict)
            return content

        return None


class TestCase(BaseTestCase):
    """Base test case for server views."""

    scenario: scenarios.DefaultContextAPI

    def get_debusine(self) -> Debusine:
        """Get a Debusine client that can query the test server."""
        with mock.patch(
            "debusine.client.debusine.DebusineHttpClient",
            partial(DjangoDebusineHttpClient, self.client),
        ):
            return Debusine(
                "/api",
                self.scenario.user_token.key,
                self.scenario.scope.name,
                logger=logging.getLogger("debusine"),
            )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Test custom path converters for package archives."""

import datetime as dt

from django.http import HttpResponse
from django.test import override_settings
from django.urls import path, resolve, reverse

# Make sure the snapshot converter is registered.
import debusine.web.archives.converters  # noqa: F401
from debusine.test.django import TestCase

urlpatterns = [
    path(
        "<str:scope>/<str:workspace>/<snapshot:snapshot>/",
        HttpResponse,
        name="archive",
    )
]


@override_settings(ROOT_URLCONF="debusine.web.archives.tests.test_converters")
class SnapshotConverterTests(TestCase):
    """Test SnapshotConverter."""

    def test_round_trip(self) -> None:
        url = "/test-scope/test-workspace/20250501T000000Z/"
        match = resolve(url)
        self.assertEqual(match.url_name, "archive")
        self.assertEqual(
            match.kwargs,
            {
                "scope": "test-scope",
                "workspace": "test-workspace",
                "snapshot": dt.datetime(2025, 5, 1, tzinfo=dt.UTC),
            },
        )
        self.assertEqual(reverse("archive", kwargs=match.kwargs), url)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Server-side task to generate indexes for a suite."""

import lzma
import shutil
import tempfile
from dataclasses import dataclass
from functools import cache, cached_property
from pathlib import Path, PurePath
from typing import Any, override

from debian.deb822 import Release
from django.contrib.postgres.aggregates import StringAgg
from django.db.backends.base.base import BaseDatabaseWrapper
from django.db.models import (
    Case,
    CharField,
    Func,
    JSONField,
    OuterRef,
    QuerySet,
    Subquery,
    Value,
    When,
)
from django.db.models.expressions import RawSQL
from django.db.models.fields.json import KT
from django.db.models.functions import Cast, Coalesce, Concat, Left
from django.db.models.sql.compiler import SQLCompiler
from django.db.models.sql.constants import INNER
from django.db.models.sql.datastructures import BaseTable
from django_pglocks import advisory_lock

from debusine.artifacts import RepositoryIndex
from debusine.artifacts.models import ArtifactCategory, WorkRequestResults
from debusine.db.locks import LockError, LockType
from debusine.db.models import (
    Artifact,
    ArtifactRelation,
    CollectionItem,
    FileInArtifact,
)
from debusine.db.orm_utils import (
    CompareBoolean,
    Encode,
    JSONBEach,
    JSONBEachText,
    JSONBToRecordSet,
    JSONBTypeOf,
    JSONStringValue,
)
from debusine.db.tasks.inputs import CollectionInput
from debusine.server.tasks import BaseServerTask
from debusine.server.tasks.models import (
    GenerateSuiteIndexesData,
    GenerateSuiteIndexesDynamicData,
)
from debusine.tasks import TaskConfigError
from debusine.utils import calculate_hash

# Loosely based on the metadata_keys table in Debian's dak database.
_metadata_ordering = (
    "Package",
    "Source",
    "Binary",
    "Version",
    "Essential",
    "Installed-Size",
    "Maintainer",
    "Uploaders",
    "Original-Maintainer",
    "Build-Depends",
    "Build-Depends-Indep",
    "Build-Depends-Arch",
    "Build-Conflicts",
    "Build-Conflicts-Indep",
    "Build-Conflicts-Arch",
    "Architecture",
    "Standards-Version",
    "Format",
    "Files",
    "Dm-Upload-Allowed",
    "Vcs-Browser",
    "Vcs-%",
    "Checksums-%",
    "Replaces",
    "Provides",
    "Depends",
    "Pre-Depends",
    "Recommends",
    "Suggests",
    "Enhances",
    "Conflicts",
    "Breaks",
    "Description",
    "Origin",
    "Bugs",
    "Multi-Arch",
    "Homepage",
    "%",
    "Tag",
    "Package-Type",
    "Installer-Menu-Item",
)


def _release_sort_key(key: str) -> tuple[int, str | None]:
    """
    Sort order for Release fields.

    This is based on what dak produces, and is cosmetic.
    """
    fixed_start = (
        "Origin",
        "Label",
        "Suite",
        "Version",
        "Codename",
        "Changelogs",
        "Date",
        "Valid-Until",
        "NotAutomatic",
        "ButAutomaticUpgrades",
        "Acquire-By-Hash",
        "No-Support-for-Architecture-all",
        "Architectures",
        "Components",
        "Description",
    )
    fixed_end = ("MD5Sum", "SHA1", "SHA256")
    try:
        return fixed_start.index(key), None
    except ValueError:
        pass
    try:
        return len(fixed_start) + 1 + fixed_end.index(key), None
    except ValueError:
        pass
    # Sort anything else alphabetically between the keys in fixed_start and
    # those in fixed_end.
    return len(fixed_start), key


class FunctionJoin(BaseTable):
    """
    Join a set-returning function.

    See:
    https://forum.djangoproject.com/t/joining-a-set-returning-function-and-aggregating-its-output/40930

    Although this is really a join, ``Query.join`` accepts either ``Join``
    or ``BaseTable``, and we subclass ``BaseTable`` instead since it's less
    inconvenient.
    """

    def __init__(
        self,
        table_name: str,
        alias: str | None,
        function: Func,
        use_alias: bool = True,
    ) -> None:
        """Construct the join."""
        super().__init__(table_name=table_name, alias=alias)
        self.function = function
        self.use_alias = use_alias

    @override
    def as_sql(
        self,
        compiler: SQLCompiler,
        connection: BaseDatabaseWrapper,  # noqa: ARG002, U100
    ) -> tuple[str, list[str | int]]:
        """Generate an SQL fragment to be included in the current query."""
        alias_str = (
            f" {self.table_name}"
            if self.table_alias is None or not self.use_alias
            else f" {self.table_alias}"
        )
        sql, params = compiler.compile(self.function)
        return f"{INNER} {sql}{alias_str} ON true", list(params)

    @override
    def relabeled_clone(
        self, change_map: dict[str | None, str]
    ) -> "FunctionJoin":
        """Return a clone of ``self``, with any column aliases relabeled."""
        return self.__class__(
            self.table_name,
            change_map.get(self.table_alias, self.table_alias),
            self.function,
            use_alias=self.use_alias,
        )

    @property
    def identity(self) -> tuple[Any, ...]:
        """Return object identity, for comparison and hashing."""
        return (
            self.__class__,
            self.table_name,
            self.table_alias,
            # https://github.com/typeddjango/django-stubs/pull/2685
            *getattr(self.function, "identity"),
        )


@dataclass
class IndexInfo:
    """Information about an index file, used to generate ``Release``."""

    size: int
    sha256: str
    artifact: Artifact | None = None


class GenerateSuiteIndexes(
    BaseServerTask[GenerateSuiteIndexesData, GenerateSuiteIndexesDynamicData]
):
    """Task that generates indexes for a suite."""

    TASK_VERSION = 1

    suite = CollectionInput(field="suite_collection")

    @override
    def post_init(self) -> None:
        super().post_init()
        self.index_info: dict[PurePath, IndexInfo] = {}
        self.indexes: dict[PurePath, Artifact] = {}

    @override
    def compute_dynamic_data(self) -> GenerateSuiteIndexesDynamicData:
        """Resolve lookups for this task."""
        dynamic_data = GenerateSuiteIndexesDynamicData(
            suite_collection_id=self.suite.collection_id
        )
        self.save_field_values(dynamic_data)
        return dynamic_data

    @cached_property
    def components(self) -> list[str]:
        """The components for which to generate indexes in this suite."""
        try:
            components = self.suite.data["components"]
        except KeyError:
            raise TaskConfigError(
                f"Collection '{self.suite.lookup}' has no "
                f"'components' configuration"
            )
        else:
            # TODO: This should move to Pydantic.
            if not isinstance(components, list):
                raise TaskConfigError(
                    f"Expected 'components' as a list; got {components!r}"
                )
            return components

    @cached_property
    def architectures(self) -> list[str]:
        """The architectures for which to generate indexes in this suite."""
        try:
            architectures = self.suite.data["architectures"]
        except KeyError:
            raise TaskConfigError(
                f"Collection '{self.data.suite_collection}' has no "
                f"'architectures' configuration"
            )
        else:
            # TODO: This should move to Pydantic.
            if not isinstance(architectures, list):
                raise TaskConfigError(
                    f"Expected 'architectures' as a list; got {architectures!r}"
                )
            return architectures

    def _add_artifact(
        self,
        *,
        path_in_suite: PurePath,
        path: Path,
        content_type: str | None = None,
        newer_item: CollectionItem | None,
    ) -> Artifact:
        artifact = Artifact.objects.create_from_local_artifact(
            RepositoryIndex.create(
                file=path,
                path=path_in_suite.as_posix(),
                content_type=content_type,
            ),
            self.workspace,
            created_by_work_request=self.work_request,
        )
        self.suite.collection.manager.add_artifact(
            artifact,
            user=self.work_request.created_by,
            workflow=self.work_request.parent,
            variables={"path": path_in_suite.as_posix()},
            name=f"index:{path_in_suite.as_posix()}",
            created_at=self.data.generate_at,
            replaced_by=newer_item,
        )
        return artifact

    def _add_index_info(
        self,
        *,
        path_in_suite: PurePath,
        path: Path,
        artifact: Artifact | None = None,
    ) -> None:
        self.index_info[path_in_suite] = IndexInfo(
            size=path.stat().st_size,
            sha256=calculate_hash(path, "sha256").hex(),
            artifact=artifact,
        )

    @cache
    def _make_field_ordering(self, key_field: RawSQL) -> Case:
        conditions: list[When] = []
        for i, name in enumerate(_metadata_ordering):
            if name == "%":
                continue
            elif "%" in name:
                conditions.append(
                    When(CompareBoolean(key_field, "LIKE", Value(name)), then=i)
                )
            else:
                conditions.append(
                    When(CompareBoolean(key_field, "=", Value(name)), then=i)
                )
        return Case(*conditions, default=_metadata_ordering.index("%"))

    def generate_sources(
        self,
        *,
        temp_path: Path,
        component: str,
        newer_item: CollectionItem | None = None,
    ) -> None:
        """Generate a ``Sources`` file."""
        qs: QuerySet[CollectionItem] = (
            CollectionItem.objects.artifacts_in_collection(
                parent_collection=self.suite.collection,
                category=ArtifactCategory.SOURCE_PACKAGE,
            )
            .active_at(self.data.generate_at)
            .annotate(component=KT("data__component"))
            .filter(
                component=component,
                # Redundant with the collection item category filter above,
                # but serves to tell Django that it needs to join
                # db_artifact before our manual join of dsc_fields.
                artifact__category=ArtifactCategory.SOURCE_PACKAGE,
            )
            # TODO: Once we have the debversion extension, we should sort by
            # true version ordering rather than just lexicographically by
            # version.
            .order_by(KT("data__package"), KT("data__version"))
        )
        qs.query.join(
            FunctionJoin(
                "dsc_fields",
                None,
                JSONBEach(RawSQL("db_artifact.data->'dsc_fields'", ())),
            )
        )

        checksums_sha256 = (
            FileInArtifact.objects.filter(artifact=OuterRef("artifact__id"))
            # See: https://stackoverflow.com/a/64902200
            .annotate(dummy_group_by=Value(1))
            .values("dummy_group_by")
            .annotate(
                field=Concat(
                    Value("Checksums-Sha256:"),
                    StringAgg(
                        Concat(
                            Value("\n "),
                            Encode("file__sha256", Value("hex")),
                            Value(" "),
                            Cast("file__size", CharField()),
                            Value(" "),
                            "path",
                        ),
                        delimiter="",
                        ordering="path",
                    ),
                    output_field=CharField(),
                )
            )
            .values("field")
        )

        # python-debian >= 0.1.50 parses Package-List, and if the artifact
        # was created with that version then that will be stored in
        # dsc_fields.  Make a subquery to serialize it again.
        package_list: QuerySet[Artifact, Any] = Artifact.objects.filter(
            id=OuterRef("artifact__id")
        )
        package_list.query.join(
            FunctionJoin(
                'package_list(package text, "package-type" text, '
                'section text, priority text, _other text)',
                None,
                JSONBToRecordSet(
                    RawSQL("db_artifact.data->'dsc_fields'->'Package-List'", ())
                ),
                use_alias=False,
            )
        )
        package_list_fields = {
            name: RawSQL(f'package_list."{name}"', (), output_field=CharField())
            for name in (
                "package",
                "package-type",
                "section",
                "priority",
                "_other",
            )
        }
        package_list = (
            package_list
            # See: https://stackoverflow.com/a/64902200
            .annotate(dummy_group_by=Value(1))
            .values("dummy_group_by")
            .annotate(
                field=Concat(
                    Value("Package-List: "),
                    StringAgg(
                        Concat(
                            Value("\n "),
                            package_list_fields["package"],
                            Value(" "),
                            package_list_fields["package-type"],
                            Value(" "),
                            package_list_fields["section"],
                            Value(" "),
                            package_list_fields["priority"],
                            Value(" "),
                            package_list_fields["_other"],
                        ),
                        delimiter="",
                    ),
                    output_field=CharField(),
                )
            )
            .values("field")
        )

        dsc_fields_key = RawSQL("dsc_fields.key", (), output_field=CharField())
        dsc_fields_value = RawSQL(
            "dsc_fields.value", (), output_field=JSONField()
        )
        render_dsc_field = Case(
            When(
                CompareBoolean(dsc_fields_key, "=", Value("Source")),
                then=Concat(
                    Value("Package: "), JSONStringValue(dsc_fields_value)
                ),
            ),
            # Debusine doesn't currently store MD5 or SHA-1 checksums.
            # Since dak generates a number of *-updates suites without those
            # (although unstable currently still has MD5 checksums), we can
            # probably get away without them.
            When(
                CompareBoolean(dsc_fields_key, "=", Value("Files")), then=None
            ),
            When(
                CompareBoolean(dsc_fields_key, "=", Value("Checksums-Sha256")),
                then=Subquery(checksums_sha256),
            ),
            When(
                CompareBoolean(dsc_fields_key, "LIKE", Value("Checksums-%")),
                then=None,
            ),
            When(
                CompareBoolean(dsc_fields_key, "=", Value("Package-List"))
                & CompareBoolean(
                    JSONBTypeOf(dsc_fields_value), "=", Value("array")
                ),
                then=Subquery(package_list),
            ),
            default=Concat(
                dsc_fields_key, Value(": "), JSONStringValue(dsc_fields_value)
            ),
        )

        qs = qs.values_list(
            Concat(
                # Emit fields from .dsc or related files.
                StringAgg(
                    render_dsc_field,
                    delimiter="\n",
                    ordering=self._make_field_ordering(dsc_fields_key),
                    output_field=CharField(),
                ),
                # Emit Directory.
                Concat(
                    Value(f"\nDirectory: pool/{component}/"),
                    Case(
                        When(
                            data__package__startswith="lib",
                            then=Left(KT("data__package"), 4),
                        ),
                        default=Left(KT("data__package"), 1),
                    ),
                    Value("/"),
                    Cast(KT("data__package"), CharField()),
                ),
                # Emit Priority.  (Debusine doesn't currently track priority
                # for source packages; dak at least sometimes uses
                # "Priority: source", which should be good enough.)
                Value("\nPriority: source"),
                # Emit Section.
                Concat(
                    Value("\nSection: "),
                    Coalesce(
                        Cast(KT("data__section"), CharField()), Value("misc")
                    ),
                ),
            ),
            flat=True,
        )

        sources_path_in_suite = PurePath(component, "source", "Sources")
        sources_path = temp_path / sources_path_in_suite
        sources_path.parent.mkdir(parents=True, exist_ok=True)
        with sources_path.open("w") as f:
            for stanza in qs:
                print(stanza, file=f)
                print(file=f)
        self._add_index_info(
            path_in_suite=sources_path_in_suite, path=sources_path
        )

        sources_xz_path_in_suite = sources_path_in_suite.with_suffix(".xz")
        sources_xz_path = sources_path.with_suffix(".xz")
        with (
            sources_path.open(mode="rb") as uncompressed,
            lzma.open(
                sources_xz_path, mode="wb", format=lzma.FORMAT_XZ
            ) as compressed,
        ):
            shutil.copyfileobj(uncompressed, compressed)
        sources_xz_artifact = self._add_artifact(
            path_in_suite=sources_xz_path_in_suite,
            path=sources_xz_path,
            newer_item=newer_item,
        )
        self._add_index_info(
            path_in_suite=sources_xz_path_in_suite,
            path=sources_xz_path,
            artifact=sources_xz_artifact,
        )

    def generate_packages(
        self,
        *,
        temp_path: Path,
        component: str,
        architecture: str,
        newer_item: CollectionItem | None = None,
    ) -> None:
        """Generate a ``Packages`` file."""
        architectures = {architecture}
        if architecture != "all" and (
            "all" not in self.architectures
            or self.suite.data.get("duplicate_architecture_all", False)
        ):
            architectures.add("all")
        qs: QuerySet[CollectionItem] = (
            CollectionItem.objects.artifacts_in_collection(
                parent_collection=self.suite.collection,
                category=ArtifactCategory.BINARY_PACKAGE,
            )
            .active_at(self.data.generate_at)
            .annotate(
                component=KT("data__component"),
                architecture=KT("data__architecture"),
            )
            .filter(
                component=component,
                architecture__in=architectures,
                # Redundant with the collection item category filter above,
                # but serves to tell Django that it needs to join
                # db_artifact before our manual join of deb_fields.
                artifact__category=ArtifactCategory.BINARY_PACKAGE,
            )
            # TODO: Once we have the debversion extension, we should sort by
            # true version ordering rather than just lexicographically by
            # version.
            .order_by(
                KT("data__srcpkg_name"),
                KT("data__package"),
                KT("data__version"),
            )
        )
        qs.query.join(
            FunctionJoin(
                "deb_fields",
                None,
                JSONBEachText(RawSQL("db_artifact.data->'deb_fields'", ())),
            )
        )
        deb_fields_key = RawSQL("deb_fields.key", (), output_field=CharField())
        deb_fields_value = RawSQL(
            "deb_fields.value", (), output_field=CharField()
        )
        render_deb_field = Case(
            When(
                CompareBoolean(deb_fields_key, "=", Value("Section")), then=None
            ),
            When(
                CompareBoolean(deb_fields_key, "=", Value("Priority")),
                then=None,
            ),
            default=Concat(deb_fields_key, Value(": "), deb_fields_value),
        )
        qs = qs.values_list(
            Concat(
                # Emit fields from .deb or related files.
                StringAgg(
                    render_deb_field,
                    delimiter="\n",
                    ordering=self._make_field_ordering(deb_fields_key),
                    output_field=CharField(),
                ),
                # Emit Section.
                Coalesce(
                    Concat(
                        Value("\nSection: "),
                        Cast(KT("data__section"), CharField()),
                    ),
                    Value(""),
                ),
                # Emit Priority.
                Coalesce(
                    Concat(
                        Value("\nPriority: "),
                        Cast(KT("data__priority"), CharField()),
                    ),
                    Value(""),
                ),
                # Emit Filename.
                # Note that debian:binary-package artifacts are guaranteed
                # to contain exactly one file.
                Concat(
                    Value(f"\nFilename: pool/{component}/"),
                    Case(
                        When(
                            data__srcpkg_name__startswith="lib",
                            then=Left(KT("data__srcpkg_name"), 4),
                        ),
                        default=Left(KT("data__srcpkg_name"), 1),
                    ),
                    Value("/"),
                    Cast(KT("data__srcpkg_name"), CharField()),
                    Value("/"),
                    "artifact__fileinartifact__path",
                ),
                # Emit Size.
                Concat(
                    Value("\nSize: "),
                    Cast("artifact__files__size", CharField()),
                ),
                # Emit SHA256.
                # Debusine doesn't currently store MD5 or SHA-1 checksums.
                # Since dak generates a number of *-updates suites without
                # those (although unstable currently still has MD5
                # checksums), we can probably get away without them.
                Concat(
                    Value("\nSHA256: "),
                    Encode("artifact__files__sha256", Value("hex")),
                ),
            ),
            flat=True,
        )

        packages_path_in_suite = PurePath(
            component, f"binary-{architecture}", "Packages"
        )
        packages_path = temp_path / packages_path_in_suite
        packages_path.parent.mkdir(parents=True, exist_ok=True)
        with packages_path.open("w") as f:
            for stanza in qs:
                print(stanza, file=f)
                print(file=f)
        self._add_index_info(
            path_in_suite=packages_path_in_suite, path=packages_path
        )

        packages_xz_path_in_suite = packages_path_in_suite.with_suffix(".xz")
        packages_xz_path = packages_path.with_suffix(".xz")
        with (
            packages_path.open(mode="rb") as uncompressed,
            lzma.open(
                packages_xz_path, mode="wb", format=lzma.FORMAT_XZ
            ) as compressed,
        ):
            shutil.copyfileobj(uncompressed, compressed)
        packages_xz_artifact = self._add_artifact(
            path_in_suite=packages_xz_path_in_suite,
            path=packages_xz_path,
            newer_item=newer_item,
        )
        self._add_index_info(
            path_in_suite=packages_xz_path_in_suite,
            path=packages_xz_path,
            artifact=packages_xz_artifact,
        )

    def generate_release(
        self, *, temp_path: Path, newer_item: CollectionItem | None = None
    ) -> None:
        """Generate a ``Release`` file."""
        release = Release()

        # Defaults; may be overridden by release_fields.
        release["Suite"] = release["Codename"] = self.suite.name

        release_fields = dict(self.suite.data.get("release_fields", {}))
        release_fields.setdefault("Acquire-By-Hash", True)

        # Convert boolean fields.
        for field in (
            "NotAutomatic",
            "ButAutomaticUpgrades",
            "Acquire-By-Hash",
        ):
            value = release_fields.get(field)
            if isinstance(value, bool):
                if value:
                    release_fields[field] = "yes"
                else:
                    del release_fields[field]

        release.update(release_fields)

        # Set or remove some fields where we're authoritative, overriding
        # anything that might be in release_fields.
        release["Date"] = self.data.generate_at.strftime(
            "%a, %d %b %Y %H:%M:%S UTC"
        )
        # TODO: Support this by adding a field to the suite specifying the
        # validity period, relative to Date.
        release.pop("Valid-Until", None)
        if self.suite.data.get("duplicate_architecture_all", False):
            release["No-Support-for-Architecture-all"] = "Packages"
        else:
            release.pop("No-Support-for-Architecture-all", None)
        release["Architectures"] = " ".join(self.suite.data["architectures"])
        release["Components"] = " ".join(self.suite.data["components"])
        release.pop("MD5Sum", None)
        release.pop("SHA1", None)
        release.pop("SHA512", None)

        # Debusine currently only stores SHA256 checksums, so for now that's
        # all that we emit.
        # TODO: This currently only handles the files that we generate every
        # time (Packages/Sources).  Some files such as Contents may need to
        # be generated on different schedules.
        related: list[Artifact] = []
        # Deb822.__setitem__'s type for value doesn't quite match how
        # Release works.
        release["SHA256"] = []  # type: ignore[assignment]
        for rel_path, index_info in sorted(self.index_info.items()):
            release["SHA256"].append(
                {
                    "sha256": index_info.sha256,
                    "size": index_info.size,
                    "name": rel_path.as_posix(),
                }
            )
            if index_info.artifact is not None:
                related.append(index_info.artifact)

        release.sort_fields(key=_release_sort_key)
        release_path_in_suite = PurePath("Release")
        release_path = temp_path / release_path_in_suite
        with release_path.open("wb") as f:
            release.dump(f)
        release_artifact = self._add_artifact(
            path_in_suite=release_path_in_suite,
            path=release_path,
            newer_item=newer_item,
            content_type="text/plain; charset=utf-8",
        )
        for target in related:
            ArtifactRelation.objects.create(
                artifact=release_artifact,
                target=target,
                type=ArtifactRelation.Relations.RELATES_TO,
            )

    def _find_indexes(self) -> QuerySet[CollectionItem]:
        """Find index items in the suite collection."""
        return self.suite.collection.child_items.filter(
            child_type=CollectionItem.Types.ARTIFACT,
            name__regex=(
                r"^index:([^/]*/[^/]*/(Sources|Packages)[^/]*|Release)$"
            ),
        )

    def supersede_older_items(self) -> None:
        """Remove old items superseded by this task."""
        previous_generation = (
            self._find_indexes()
            .filter(
                created_at__lt=self.data.generate_at,
            )
            .order_by("-created_at")
            .first()
        )
        if previous_generation is not None:
            self._find_indexes().filter(
                created_at=previous_generation.created_at,
            ).update(
                removed_by_user=self.work_request.created_by,
                removed_by_workflow=self.work_request.parent,
                removed_at=self.data.generate_at,
            )

    def find_newer_item(self) -> CollectionItem | None:
        """Find the earliest item newer than the ones we're generating."""
        return (
            self._find_indexes()
            .filter(
                created_at__gt=self.data.generate_at,
            )
            # Fall back to ordering by ID, to improve testability.
            .order_by("created_at", "id")
            .first()
        )

    @override
    def _execute(self) -> WorkRequestResults:
        """Execute the task."""
        with advisory_lock(
            (
                LockType.GENERATE_SUITE_INDEXES,
                # Only use the bottom 31 bits, in order that this fits into
                # PostgreSQL's int type.  In the unlikely event that we have
                # enough mirrored collections for there to be a collision,
                # then it just means that the colliding collections can't
                # have indexes generated for them simultaneously.
                self.suite.collection_id & (2**31 - 1),
            ),
            wait=False,
        ) as acquired:
            if not acquired:
                raise LockError(
                    f"Another GenerateSuiteIndexes task for "
                    f"{self.data.suite_collection} is already running"
                )

            self.indexes = {}
            self.supersede_older_items()
            newer_item = self.find_newer_item()
            with tempfile.TemporaryDirectory(
                prefix="debusine-generate-suite-indexes-"
            ) as temp_dir:
                temp_path = Path(temp_dir)
                for component in self.components:
                    self.generate_sources(
                        temp_path=temp_path,
                        component=component,
                        newer_item=newer_item,
                    )
                    for architecture in self.architectures:
                        self.generate_packages(
                            temp_path=temp_path,
                            component=component,
                            architecture=architecture,
                            newer_item=newer_item,
                        )
                self.generate_release(
                    temp_path=temp_path, newer_item=newer_item
                )
                return WorkRequestResults.SUCCESS

    @override
    def get_label(self) -> str:
        """Return the task label."""
        return (
            f"generate indexes for {self.data.suite_collection} "
            f"at {self.data.generate_at}"
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the worker views."""

import datetime as dt
import logging
import secrets
from collections.abc import Generator
from typing import Any, ClassVar, override
from unittest import mock

import pytest
from django.db import OperationalError, transaction
from django.db.models import Max
from django.test import override_settings
from django.urls import reverse
from django.utils import timezone
from rest_framework import status

from debusine.artifacts.models import RuntimeStatistics
from debusine.db.models import Token, WorkRequest, Worker
from debusine.db.models.work_requests import (
    MAX_AUTOMATIC_RETRIES,
    WorkRequestQuerySet,
)
from debusine.db.tests.utils import RunInParallelTransaction, raise_db_exception
from debusine.project.celery import celery_tasks_in_progress
from debusine.server.views.rest import (
    IsWorkRequestTokenAuthenticated,
    IsWorkerTokenAuthenticated,
)
from debusine.server.views.tests.test_work_requests import WorkRequestTestCase
from debusine.server.views.workers import (
    GetNextWorkRequestView,
    UpdateWorkRequestAsCompletedView,
    UpdateWorkerDynamicMetadataView,
    update_work_request_as_completed,
)
from debusine.server.workflows.models import WorkRequestWorkflowData
from debusine.tasks.models import (
    OutputData,
    OutputDataError,
    SbuildData,
    SbuildInput,
    WorkerType,
)
from debusine.test.django import TestCase, TransactionTestCase


class RegisterViewTests(TestCase):
    """Tests for the RegisterView class."""

    def test_create_token_and_worker(self) -> None:
        """Token and Worker are created by the view."""
        self.assertQuerySetEqual(Token.objects.all(), [])
        self.assertQuerySetEqual(Worker.objects.all(), [])

        time_start = timezone.now()

        token_key = secrets.token_hex(32)

        data = {'token': token_key, 'fqdn': 'worker-bee.lan'}

        response = self.client.post(reverse('api:register'), data)

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        token = Token.objects.get_token_or_none(token_key=token_key)
        assert token is not None

        worker = token.worker

        # Assert created worker
        self.assertEqual(worker.name, 'worker-bee-lan')
        self.assertGreaterEqual(worker.registered_at, time_start)
        self.assertLessEqual(worker.registered_at, timezone.now())
        self.assertIsNone(worker.connected_at)
        self.assertEqual(worker.worker_type, WorkerType.EXTERNAL)

        # Assert created token
        self.assertGreaterEqual(token.created_at, time_start)
        self.assertLessEqual(token.created_at, timezone.now())
        self.assertIsNone(token.user)
        self.assertEqual(token.comment, '')
        self.assertFalse(token.enabled)

    def test_create_token_and_worker_duplicate_name(self) -> None:
        """Token is created and Worker disambiguated if needed."""
        token_1 = self.playground.create_worker_token(set_worker=False)
        Worker.objects.create_with_fqdn('worker-lan', token=token_1)

        token_key = secrets.token_hex(32)
        data = {'token': token_key, 'fqdn': 'worker.lan'}
        response = self.client.post(reverse('api:register'), data)

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertTrue(Worker.objects.filter(name='worker-lan-2').exists())

    def test_create_token_and_worker_signing(self) -> None:
        """Register a signing worker."""
        self.assertQuerySetEqual(Token.objects.all(), [])
        self.assertQuerySetEqual(Worker.objects.all(), [])

        time_start = timezone.now()

        token_key = secrets.token_hex(32)

        data = {
            'token': token_key,
            'fqdn': 'worker-bee.lan',
            'worker_type': WorkerType.SIGNING,
        }

        response = self.client.post(reverse('api:register'), data)

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        token = Token.objects.get_token_or_none(token_key=token_key)
        assert token is not None

        worker = token.worker

        # Assert created worker
        self.assertEqual(worker.name, 'worker-bee-lan')
        self.assertGreaterEqual(worker.registered_at, time_start)
        self.assertLessEqual(worker.registered_at, timezone.now())
        self.assertIsNone(worker.connected_at)
        self.assertEqual(worker.worker_type, WorkerType.SIGNING)

        # Assert created token
        self.assertGreaterEqual(token.created_at, time_start)
        self.assertLessEqual(token.created_at, timezone.now())
        self.assertIsNone(token.user)
        self.assertEqual(token.comment, '')

    def test_invalid_data(self) -> None:
        """Request is refused if we have bad data."""
        data = {
            'token': secrets.token_hex(128),  # Too long
            'fqdn': 'worker.lan',
        }

        response = self.client.post(reverse('api:register'), data)

        self.assertResponseProblem(
            response,
            "Cannot deserialize worker",
            validation_errors_pattern=(
                r"Ensure this field has no more than .* characters"
            ),
        )
        self.assertIsInstance(response.json()["validation_errors"], dict)

        self.assertFalse(Worker.objects.filter(name='worker-lan').exists())

    def test_activation_token(self) -> None:
        """Register using an activation token."""
        activation_token = self.playground.create_worker_activation_token()
        worker = activation_token.activating_worker
        assert worker is not None
        token_key = secrets.token_hex(32)
        data = {"token": token_key, "fqdn": "worker.lan"}

        response = self.client.post(
            reverse("api:register"),
            data,
            headers={"token": activation_token.key},
        )

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        worker.refresh_from_db()
        self.assertIsNone(worker.activation_token)
        token = Token.objects.get_token_or_none(token_key=token_key)
        assert token is not None
        self.assertEqual(token.worker, worker)
        self.assertTrue(token.enabled)
        self.assertFalse(
            Token.objects.get_token_or_none(token_key=activation_token.key)
        )

    def test_invalid_activation_token(self) -> None:
        """An invalid activation token is rejected outright."""
        token_key = secrets.token_hex(32)
        data = {"token": token_key, "fqdn": "worker.lan"}

        response = self.client.post(
            reverse("api:register"), data, headers={"token": token_key}
        )

        self.assertResponseProblem(
            response, title="Invalid worker activation token"
        )
        self.assertFalse(Token.objects.get_token_or_none(token_key=token_key))


class GetNextWorkRequestViewTests(WorkRequestTestCase):
    """Tests for GetNextWorkRequestView."""

    token: ClassVar[Token]
    worker: ClassVar[Worker]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.worker = cls.playground.create_worker(fqdn="worker-test")
        assert cls.worker.token is not None
        cls.token = cls.worker.token

    def create_work_request(
        self, created_at: dt.datetime, **kwargs: Any
    ) -> WorkRequest:
        """Return a new work_request as specified by the parameters."""
        environment_item = self.playground.create_debian_environment()
        assert environment_item.artifact is not None
        source_artifact = self.playground.create_source_artifact()
        task_data = SbuildData(
            input=SbuildInput(source_artifact=source_artifact.id),
            build_architecture="test",
            environment=environment_item.artifact.id,
        )

        work_request = self.playground.create_worker_task(
            worker=self.worker,
            task_name="sbuild",
            task_data=task_data,
            **kwargs,
        )

        work_request.created_at = created_at

        work_request.save()

        return work_request

    def test_check_permissions(self) -> None:
        """Only authenticated requests are processed by the view."""
        self.assertIn(
            IsWorkerTokenAuthenticated,
            GetNextWorkRequestView.permission_classes,
        )

    def test_completion_in_progress(self) -> None:
        """A work request that's pending completion is left alone."""
        work_request = self.create_work_request(
            dt.datetime.now(dt.UTC) - dt.timedelta(minutes=10),
            mark_running=True,
        )
        work_request_pending = self.create_work_request(
            dt.datetime.now(dt.UTC), mark_pending=True
        )
        self.playground.create_build_log_artifact(work_request=work_request)
        self.assertFalse(
            celery_tasks_in_progress(
                "debusine.server.views.workers.update_work_request_as_completed"
            ).exists()
        )
        token = self.playground.create_work_request_token(work_request)

        with self.captureOnCommitCallbacks(execute=True):
            response = self.client.put(
                reverse(
                    "api:work-request-completed",
                    kwargs={"work_request_id": work_request.id},
                ),
                data={"result": WorkRequest.Results.SUCCESS},
                headers={"token": token.key},
                content_type="application/json",
            )

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertTrue(
            celery_tasks_in_progress(
                "debusine.server.views.workers.update_work_request_as_completed"
            ).exists()
        )

        with self.assertNoLogs(
            logger="debusine.server.views.workers", level=logging.DEBUG
        ):
            response = self.client.post(
                reverse("api:work-request-get-next"),
                headers={"token": self.token.key},
            )

        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        work_request.refresh_from_db()
        self.assertEqual(work_request.status, WorkRequest.Statuses.RUNNING)
        self.assertFalse(hasattr(work_request, "superseded"))

        # Pretend that the Celery task completed normally.
        with (
            override_settings(PGTRANSACTION_RETRY=0),
            mock.patch("pgtransaction.Atomic.execute_set_transaction_modes"),
        ):
            celery_result = update_work_request_as_completed.apply(
                args=(work_request.id,),
                kwargs={
                    "result": WorkRequest.Results.SUCCESS,
                    "output_data": None,
                },
            )
            self.assertIsNone(celery_result.get())
        celery_tasks_in_progress(
            "debusine.server.views.workers.update_work_request_as_completed"
        ).delete()

        with self.assertNoLogs(
            logger="debusine.server.views.workers", level=logging.DEBUG
        ):
            response = self.client.post(
                reverse("api:work-request-get-next"),
                headers={"token": self.token.key},
            )

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        work_request.refresh_from_db()
        self.assertEqual(work_request.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(work_request.result, WorkRequest.Results.SUCCESS)
        self.assertFalse(hasattr(work_request, "superseded"))
        self.check_response_for_work_request(
            response, work_request_pending, with_token=True
        )

    def test_get_running_work_request(self) -> None:
        """A running work request with no output is simply retried."""
        work_request = self.playground.create_worker_task(
            mark_running=True, worker=self.worker
        )

        with self.assertLogsContains(
            f"Re-attempting work request {work_request.id}",
            logger="debusine.server.views.workers",
            level=logging.DEBUG,
        ):
            response = self.client.post(
                reverse('api:work-request-get-next'),
                headers={"token": self.token.key},
            )

        # The running work request was retried
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        work_request.refresh_from_db()
        self.assertEqual(work_request.status, WorkRequest.Statuses.RUNNING)
        self.assertFalse(hasattr(work_request, "superseded"))
        self.assertEqual(work_request.workflow_data.retry_count, 1)
        self.check_response_for_work_request(
            response, work_request, with_token=True
        )

    def test_get_running_work_request_retry_limit(self) -> None:
        """A work request is not simply retried after the automatic limit."""
        work_request = self.playground.create_worker_task(
            mark_running=True,
            worker=self.worker,
            workflow_data=WorkRequestWorkflowData(
                retry_count=MAX_AUTOMATIC_RETRIES
            ),
        )

        with self.assertLogsContains(
            (
                f"Cannot retry previously-running work request "
                f"{work_request.id}: automatic retry limit exceeded."
            ),
            logger="debusine.server.views.workers",
            level=logging.DEBUG,
        ):
            response = self.client.post(
                reverse('api:work-request-get-next'),
                headers={"token": self.token.key},
            )

        # The running work request was not retried
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        work_request.refresh_from_db()
        self.assertEqual(work_request.status, WorkRequest.Statuses.ABORTED)
        self.assertEqual(
            work_request.output_data,
            OutputData(
                errors=[
                    OutputDataError(
                        message=(
                            f"No worker managed to complete work request "
                            f"{work_request.id}. Automatic retry limit "
                            f"exceeded."
                        ),
                        code="work-request-retry-limit-exceeded",
                    )
                ]
            ),
        )
        self.assertFalse(hasattr(work_request, "superseded"))

    def test_get_running_partially_completed_work_request(self) -> None:
        """A partially-completed work request is aborted and retried."""
        # Create WorkRequest pending
        work_request_pending = self.create_work_request(
            dt.datetime(2022, 1, 5, 10, 13, 20, 204242, dt.UTC),
            mark_pending=True,
        )

        # Create WorkRequest running with output
        work_request_running = self.create_work_request(
            dt.datetime(2022, 1, 5, 11, 14, 22, 242178, dt.UTC),
            mark_running=True,
        )
        self.playground.create_build_log_artifact(
            work_request=work_request_running
        )

        # Request
        with self.assertLogs(
            logger="debusine.server.views.workers", level=logging.DEBUG
        ) as log:
            response = self.client.post(
                reverse("api:work-request-get-next"),
                headers={"token": self.token.key},
            )

        self.assertEqual(
            log.output,
            [
                f"INFO:debusine.server.views.workers:"
                f"Retrying stuck work request {work_request_running.pk}"
            ],
        )

        # The running work request was retried, though not yet assigned to a
        # worker; we got the pending one.
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        work_request_running.refresh_from_db()
        self.assertEqual(
            work_request_running.status, WorkRequest.Statuses.ABORTED
        )
        self.assertTrue(hasattr(work_request_running, "superseded"))
        self.assertIsNone(work_request_running.superseded.worker)
        self.assertEqual(work_request_running.workflow_data.retry_count, 0)
        self.assertEqual(
            work_request_running.superseded.workflow_data.retry_count, 1
        )
        self.check_response_for_work_request(
            response, work_request_pending, with_token=True
        )

    def test_get_running_preserves_parent_workflow(self) -> None:
        """The parent of a partially-completed work request is undisturbed."""
        workflow = self.playground.create_workflow(mark_running=True)
        # Allow dependency failures so that aborting all children of the
        # workflow would normally mark it as completed.
        workflow.workflow_data = WorkRequestWorkflowData(
            allow_dependency_failures=True
        )
        workflow.save()

        work_request_running = workflow.create_child_worker("noop")
        self.playground.advance_work_request(
            work_request_running, mark_running=True, worker=self.worker
        )
        self.playground.create_build_log_artifact(
            work_request=work_request_running
        )

        # Request
        with self.assertLogs(
            logger="debusine.server.views.workers", level=logging.DEBUG
        ) as log:
            response = self.client.post(
                reverse("api:work-request-get-next"),
                headers={"token": self.token.key},
            )

        self.assertEqual(
            log.output,
            [
                f"INFO:debusine.server.views.workers:"
                f"Retrying stuck work request {work_request_running.pk}"
            ],
        )

        # The running work request was retried, though not yet assigned to a
        # worker.
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        work_request_running.refresh_from_db()
        self.assertEqual(
            work_request_running.status, WorkRequest.Statuses.ABORTED
        )
        self.assertTrue(hasattr(work_request_running, "superseded"))
        self.assertIsNone(work_request_running.superseded.worker)
        self.assertEqual(work_request_running.workflow_data.retry_count, 0)
        self.assertEqual(
            work_request_running.superseded.workflow_data.retry_count, 1
        )
        # The parent workflow is still running.
        workflow.refresh_from_db()
        self.assertEqual(workflow.status, WorkRequest.Statuses.RUNNING)

    def test_get_running_work_request_cannot_retry(self) -> None:
        """If a running work request cannot be retried, it is logged."""
        work_request_pending = self.create_work_request(
            dt.datetime(2022, 1, 5, 10, 13, 20, 204242, dt.UTC),
            mark_pending=True,
        )
        work_request_running = self.create_work_request(
            dt.datetime(2022, 1, 5, 11, 14, 22, 242178, dt.UTC),
            mark_running=True,
        )
        work_request_running.task_data = {"invalid": True}
        work_request_running.configured_task_data = {"invalid": True}
        work_request_running.save()
        self.playground.create_build_log_artifact(
            work_request=work_request_running
        )

        with self.assertLogs(
            logger="debusine.server.views.workers", level=logging.DEBUG
        ) as log:
            response = self.client.post(
                reverse('api:work-request-get-next'),
                headers={"token": self.token.key},
            )
            self.assertAPIResponseOk(response)

        self.assertEqual(
            log.output,
            [
                f"INFO:debusine.server.views.workers:"
                f"Retrying stuck work request {work_request_running.pk}",
                "DEBUG:debusine.server.views.workers:"
                "Cannot retry previously-running work request: "
                "Task dependencies cannot be satisfied",
            ],
        )

        # The running work request was aborted but could not be retried; we
        # got the pending one.
        work_request_running.refresh_from_db()
        self.assertEqual(
            work_request_running.status, WorkRequest.Statuses.ABORTED
        )
        self.assertFalse(hasattr(work_request_running, "superseded"))
        self.check_response_for_work_request(
            response, work_request_pending, with_token=True
        )

    def test_get_work_request_change_status(self) -> None:
        """Get a WorkRequest change the status to running."""
        work_request = self.create_work_request(
            dt.datetime(2022, 1, 5, 10, 13, 20, 204242, dt.UTC),
            mark_pending=True,
        )

        # Request
        response = self.client.post(
            reverse('api:work-request-get-next'),
            headers={"token": self.token.key},
        )

        # Assert that we got the WorkRequest running
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.check_response_for_work_request(
            response, work_request, with_token=True
        )

        work_request.refresh_from_db()

        self.assertEqual(work_request.status, WorkRequest.Statuses.RUNNING)

    def test_get_work_request_configured_task_data(self) -> None:
        """Get a WorkRequest shows configured task data."""
        work_request = self.create_work_request(
            dt.datetime(2022, 1, 5, 10, 13, 20, 204242, dt.UTC),
            mark_pending=True,
        )
        work_request.task_data = {"test": 1}
        work_request.configured_task_data = {"test": 2}
        work_request.save()

        # Request
        response = self.client.post(
            reverse('api:work-request-get-next'),
            headers={"token": self.token.key},
        )

        # Assert that we got the WorkRequest running
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.check_response_for_work_request(
            response, work_request, with_token=True
        )
        self.assertEqual(response.json()["task_data"], {"test": 2})

    def test_get_older_pending(self) -> None:
        """View return the older pending work request."""
        # Create WorkRequest pending
        work_request_pending_01 = self.create_work_request(
            dt.datetime(2022, 1, 5, 10, 13, 20, 204242, dt.UTC),
            mark_pending=True,
        )

        work_request_pending_02 = self.create_work_request(
            dt.datetime(2022, 1, 5, 10, 13, 22, 204242, dt.UTC),
            mark_pending=True,
        )
        # Request
        response = self.client.post(
            reverse('api:work-request-get-next'),
            headers={"token": self.token.key},
        )

        self.check_response_for_work_request(
            response, work_request_pending_01, with_token=True
        )

        # work_request_pending_01 status changed. Let's move it back to
        # pending. This is not an allowed transition but helps here to test
        # that clients will receive the older pending
        work_request_pending_01.status = WorkRequest.Statuses.PENDING

        # Change created_at to force a change in the order that debusine
        # sends the WorkRequests to the client
        work_request_pending_01.created_at = dt.datetime(
            2022, 1, 5, 10, 13, 24, 204242, dt.UTC
        )

        work_request_pending_01.save()

        # New request to check the new order
        response = self.client.post(
            reverse('api:work-request-get-next'),
            headers={"token": self.token.key},
        )

        self.check_response_for_work_request(
            response, work_request_pending_02, with_token=True
        )

    def test_get_no_work_request(self) -> None:
        """Assert debusine sends HTTP 204 when nothing assigned to Worker."""
        response = self.client.post(
            reverse('api:work-request-get-next'),
            headers={"token": self.token.key},
        )

        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

    def test_authentication_credentials_not_provided(self) -> None:
        """Assert Token is required to use the endpoint."""
        # This is to double-check that IsWorkerTokenAuthenticated (inclusion
        # is verified in test_check_permissions) is returning what it should do.
        # All the views with IsWorkerTokenAuthenticated behave the same way
        # IsWorkerTokenAuthenticated is tested in
        # IsWorkerTokenAuthenticatedTests
        response = self.client.post(reverse('api:work-request-get-next'))
        self.assertEqual(
            response.json(),
            {
                "title": "Error",
                "detail": "Authentication credentials were not provided.",
            },
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)


class GetNextWorkRequestViewTransactionTests(TransactionTestCase):
    """Test transactional behaviour of GetNextWorkRequestView."""

    def test_locks_workflow(self) -> None:
        """The view locks all work requests in the affected workflows."""
        with transaction.atomic():
            workflows = [self.playground.create_workflow() for _ in range(5)]
            work_requests = [
                workflow.create_child_worker("noop")
                for workflow in (
                    workflows[0],
                    workflows[0],
                    workflows[1],
                    workflows[1],
                    workflows[2],
                    workflows[3],
                )
            ]
            # workflows[0] and workflows[1] have running work requests.
            for work_request in (work_requests[0], work_requests[2]):
                self.playground.advance_work_request(
                    work_request, mark_running=True
                )
                self.playground.create_build_log_artifact(
                    work_request=work_request
                )
            assert work_requests[0].worker is not None
            assert work_requests[0].worker.token is not None
            # workflows[2] and workflows[3] have pending work requests
            # assigned to the same worker as work_requests[0], with the one
            # in workflows[2] having higher priority.
            work_requests[4].priority_adjustment += 1
            work_requests[4].save()
            for work_request in (work_requests[4], work_requests[5]):
                self.playground.advance_work_request(
                    work_request,
                    mark_pending=True,
                    worker=work_requests[0].worker,
                )

        with transaction.atomic():
            response = self.client.post(
                reverse("api:work-request-get-next"),
                headers={"token": work_requests[0].worker.token.key},
            )

            self.assertEqual(response.status_code, status.HTTP_200_OK)
            # Each work request in each of the affected workflows is locked,
            # but not unaffected work requests.
            self.assert_work_requests_locked(
                locked=[
                    workflows[0],
                    work_requests[0],
                    work_requests[1],
                    workflows[2],
                    work_requests[4],
                ],
                unlocked=[
                    workflows[1],
                    work_requests[2],
                    work_requests[3],
                    workflows[3],
                    work_requests[5],
                    workflows[4],
                ],
            )

    def test_get_work_request_locked(self) -> None:
        """If the workflow is already locked, return HTTP 423."""
        with transaction.atomic():
            workflow = self.playground.create_workflow(mark_running=True)
            work_requests = [
                workflow.create_child_worker("noop") for _ in range(2)
            ]
            self.playground.advance_work_request(
                work_requests[0], mark_pending=True, assign_new_worker=True
            )
            assert work_requests[0].worker is not None
            assert work_requests[0].worker.token is not None

        thread = RunInParallelTransaction(
            lambda: WorkRequest.objects.select_for_update().get(
                id=work_requests[1].id
            )
        )
        thread.start_transaction()

        try:
            with transaction.atomic():
                with self.assertLogsContains(
                    f"Failed to lock work requests assigned to "
                    f"{work_requests[0].worker.name}",
                    logger="debusine.server.views.workers",
                    level=logging.DEBUG,
                ):
                    response = self.client.post(
                        reverse("api:work-request-get-next"),
                        headers={"token": work_requests[0].worker.token.key},
                    )

                self.assertResponseProblem(
                    response,
                    title="Retry",
                    detail_pattern=(
                        "Could not acquire workflow locks; retry later"
                    ),
                    status_code=status.HTTP_423_LOCKED,
                )
                work_requests[0].refresh_from_db()
                self.assertEqual(
                    work_requests[0].status, WorkRequest.Statuses.PENDING
                )
        finally:
            thread.stop_transaction()


class UpdateWorkRequestAsCompletedViewTests(TransactionTestCase):
    """Tests for UpdateWorkRequestAsCompletedView class and its Celery task."""

    token: Token
    worker: Worker
    work_request: WorkRequest

    @override
    def setUp(self) -> None:
        super().setUp()
        self.worker = self.playground.create_worker(fqdn="worker-test")

        environment_item = self.playground.create_debian_environment()
        assert environment_item.artifact is not None
        source_artifact = self.playground.create_source_artifact()
        task_data = SbuildData(
            input=SbuildInput(source_artifact=source_artifact.id),
            build_architecture="test",
            environment=environment_item.artifact.id,
        )

        self.work_request = self.playground.create_worker_task(
            worker=self.worker,
            task_name='sbuild',
            task_data=task_data,
            mark_running=True,
        )
        self.token = self.playground.create_work_request_token(
            self.work_request
        )

    def test_check_permissions(self) -> None:
        """Only authenticated requests are processed by the view."""
        self.assertIn(
            IsWorkRequestTokenAuthenticated,
            UpdateWorkRequestAsCompletedView.permission_classes,
        )

    def put_api_work_request_completed(
        self, result: WorkRequest.Results, output_data: OutputData | None = None
    ) -> None:
        """Assert Worker can update WorkRequest result to completed-success."""
        self.playground.advance_work_request(
            self.work_request, mark_running=True
        )

        data: dict[str, Any] = {"result": result}
        if output_data is not None:
            data["output_data"] = output_data.model_dump(
                mode="json", exclude_unset=True
            )
        with mock.patch(
            "debusine.server.views.workers.update_work_request_as_completed"
            ".apply_async"
        ) as mock_apply_async:
            response = self.client.put(
                reverse(
                    "api:work-request-completed",
                    kwargs={"work_request_id": self.work_request.id},
                ),
                data=data,
                headers={"token": self.token.key},
                content_type="application/json",
            )

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        mock_apply_async.assert_called_once_with(
            args=(self.work_request.id,),
            kwargs={"result": result, "output_data": data.get("output_data")},
            task_id=mock.ANY,
        )

        # The work request status is not changed yet; that work is deferred
        # to a Celery task.
        self.work_request.refresh_from_db()
        self.assertEqual(self.work_request.status, WorkRequest.Statuses.RUNNING)

        celery_result = update_work_request_as_completed.apply(
            args=(self.work_request.id,),
            kwargs={
                "result": data["result"],
                "output_data": data.get("output_data"),
            },
        )

        self.assertIsNone(celery_result.get())
        self.work_request.refresh_from_db()
        self.assertEqual(
            self.work_request.status, WorkRequest.Statuses.COMPLETED
        )
        self.assertEqual(self.work_request.result, result)
        self.assertEqual(self.work_request.output_data, output_data)

    def test_update_completed_result_is_success(self) -> None:
        """Assert Worker can update WorkRequest result to completed-success."""
        self.put_api_work_request_completed(WorkRequest.Results.SUCCESS)

    def test_update_completed_result_is_success_with_output_data(self) -> None:
        """Assert Worker can update WorkRequest result with output data."""
        self.put_api_work_request_completed(
            WorkRequest.Results.SUCCESS,
            OutputData(runtime_statistics=RuntimeStatistics(duration=60)),
        )

    def test_update_completed_result_is_failure(self) -> None:
        """Assert Worker can update WorkRequest result to completed-error."""
        self.put_api_work_request_completed(WorkRequest.Results.ERROR)

    def test_update_work_request_id_not_found(self) -> None:
        """Assert API returns HTTP 404 for a non-existing WorkRequest id."""
        max_id = WorkRequest.objects.aggregate(Max('id'))['id__max']

        response = self.client.put(
            reverse(
                'api:work-request-completed',
                kwargs={'work_request_id': max_id + 1},
            ),
            data={"result": WorkRequest.Results.SUCCESS},
            headers={"token": self.token.key},
            content_type="application/json",
        )

        self.assertResponseProblem(
            response,
            "Work request not found",
            status_code=status.HTTP_404_NOT_FOUND,
        )

    def test_celery_work_request_id_not_found(self) -> None:
        """The Celery task fails if the work request does not exist."""
        max_id = WorkRequest.objects.aggregate(Max("id"))["id__max"]

        with self.assertLogsContains(
            f"Work request {max_id + 1} does not exist",
            logger="debusine.server.views.workers",
            level=logging.ERROR,
        ):
            result = update_work_request_as_completed.apply(
                args=(max_id + 1,),
                kwargs={"result": WorkRequest.Results.SUCCESS},
            )

        with self.assertRaises(WorkRequest.DoesNotExist):
            result.get()

    def test_celery_work_request_already_completed(self) -> None:
        """The Celery task fails if the work request was already completed."""
        self.work_request.mark_completed(WorkRequest.Results.SUCCESS)

        with self.assertLogsContains(
            f"Failed to mark work request {self.work_request.id} as completed",
            logger="debusine.server.views.workers",
            level=logging.ERROR,
        ):
            result = update_work_request_as_completed.apply(
                args=(self.work_request.id,),
                kwargs={"result": WorkRequest.Results.SUCCESS},
            )

        self.assertIsNone(result.get())
        self.work_request.refresh_from_db()
        self.assertFalse(hasattr(self.work_request, "superseded"))
        self.assertEqual(
            self.work_request.status, WorkRequest.Statuses.COMPLETED
        )
        self.assertEqual(self.work_request.result, WorkRequest.Results.SUCCESS)

    def test_celery_work_request_failure_when_running(self) -> None:
        """The Celery task prevents the worker getting stuck on failure."""
        self.playground.advance_work_request(
            self.work_request, mark_running=True
        )

        with (
            self.assertLogsContains(
                f"Failed to mark work request {self.work_request.id} as"
                f" completed",
                logger="debusine.server.views.workers",
                level=logging.ERROR,
            ),
            mock.patch.object(
                WorkRequest,
                "mark_completed",
                side_effect=lambda *args, **kwargs: raise_db_exception(
                    "serialization_failure"
                ),
            ),
        ):
            result = update_work_request_as_completed.apply(
                args=(self.work_request.id,),
                kwargs={"result": WorkRequest.Results.SUCCESS},
            )

        self.assertIsNone(result.get())
        self.work_request.refresh_from_db()
        assert self.work_request.superseded is not None
        self.assertEqual(
            self.work_request.superseded.workflow_data.retry_count, 1
        )

    def test_update_failure_invalid_result(self) -> None:
        """Assert Worker get HTTP 400 and error for invalid result field."""
        self.playground.advance_work_request(
            self.work_request, mark_running=True
        )

        response = self.client.put(
            reverse(
                'api:work-request-completed',
                kwargs={'work_request_id': self.work_request.id},
            ),
            data={"result": "something not recognised"},
            headers={"token": self.token.key},
            content_type="application/json",
        )

        self.assertResponseProblem(
            response,
            "Cannot change work request as completed",
            validation_errors_pattern="something not recognised",
        )

    def test_unauthorized(self) -> None:
        """Assert worker cannot modify a task of another worker."""
        another_worker = self.playground.create_worker(
            fqdn="another-worker-test"
        )

        environment_item = self.playground.create_debian_environment()
        assert environment_item.artifact is not None
        source_artifact = self.playground.create_source_artifact()
        task_data = SbuildData(
            input=SbuildInput(source_artifact=source_artifact.id),
            build_architecture="test",
            environment=environment_item.artifact.id,
        )

        another_work_request = self.playground.create_worker_task(
            worker=another_worker,
            task_name='sbuild',
            task_data=task_data,
            mark_pending=True,
        )

        response = self.client.put(
            reverse(
                'api:work-request-completed',
                kwargs={'work_request_id': another_work_request.id},
            ),
            headers={"token": self.token.key},
            data={"result": WorkRequest.Results.SUCCESS},
            content_type="application/json",
        )

        self.assertResponseProblem(
            response,
            "Work request not found",
            status_code=status.HTTP_404_NOT_FOUND,
        )

    def test_locks_workflow(self) -> None:
        """The Celery task locks all work requests in the workflow."""
        with transaction.atomic():
            workflow = self.playground.create_workflow()
            work_requests = [
                workflow.create_child_worker("noop") for _ in range(2)
            ]
            self.playground.advance_work_request(
                work_requests[0], mark_running=True
            )
            assert work_requests[0].worker is not None
            assert work_requests[0].worker.token is not None
            unrelated = self.playground.create_worker_task()

        with (
            override_settings(PGTRANSACTION_RETRY=0),
            mock.patch("pgtransaction.Atomic.execute_set_transaction_modes"),
            transaction.atomic(),
        ):
            result = update_work_request_as_completed.apply(
                args=(work_requests[0].id,),
                kwargs={"result": WorkRequest.Results.SUCCESS},
            )
            self.assertIsNone(result.get())

            work_requests[0].refresh_from_db()
            self.assertEqual(
                work_requests[0].status, WorkRequest.Statuses.COMPLETED
            )
            self.assertEqual(
                work_requests[0].result, WorkRequest.Results.SUCCESS
            )
            self.assert_work_requests_locked(
                locked=[workflow, *work_requests], unlocked=[unrelated]
            )

    def test_deadlocks(self) -> None:
        """The Celery task retries automatically if it deadlocks."""
        with transaction.atomic():
            workflow = self.playground.create_workflow()
            work_requests = [
                workflow.create_child_worker("noop") for _ in range(2)
            ]
            for work_request in work_requests:
                self.playground.advance_work_request(
                    work_request, mark_running=True
                )

        def lock_workflows_side_effect() -> Generator[Any]:
            with pytest.raises(OperationalError) as exc_info:
                raise_db_exception("deadlock_detected")
            yield exc_info.value

            while True:
                yield mock.DEFAULT

        with self.spy_object(
            WorkRequestQuerySet,
            "lock_workflows_for_update",
            side_effect=lock_workflows_side_effect(),
        ) as mock_lock_workflows_for_update:
            result = update_work_request_as_completed.apply(
                args=(work_requests[0].id,),
                kwargs={"result": WorkRequest.Results.SUCCESS},
            )
            self.assertIsNone(result.get())

        self.assertEqual(mock_lock_workflows_for_update.call_count, 2)
        work_requests[0].refresh_from_db()
        self.assertEqual(
            work_requests[0].status, WorkRequest.Statuses.COMPLETED
        )
        self.assertEqual(work_requests[0].result, WorkRequest.Results.SUCCESS)


class UpdateWorkerDynamicMetadataViewTests(TestCase):
    """Tests for UpdateWorkerDynamicMetadataView class."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.worker_01 = Worker.objects.create_with_fqdn(
            "worker-01-lan",
            token=self.playground.create_worker_token(set_worker=False),
        )
        self.worker_02 = Worker.objects.create_with_fqdn(
            "worker-02-lan",
            token=self.playground.create_worker_token(set_worker=False),
        )

    def test_check_permissions(self) -> None:
        """Only authenticated requests are processed by the view."""
        self.assertIn(
            IsWorkerTokenAuthenticated,
            UpdateWorkerDynamicMetadataView.permission_classes,
        )

    def test_update_metadata_success(self) -> None:
        """Worker's dynamic_metadata is updated."""
        assert self.worker_01.token is not None
        self.assertEqual(self.worker_01.dynamic_metadata, {})
        self.assertEqual(self.worker_02.dynamic_metadata, {})

        metadata = {"cpu_cores": 4, "ram": 16}
        response = self.client.put(
            reverse('api:worker-dynamic-metadata'),
            data=metadata,
            headers={"token": self.worker_01.token.key},
            content_type="application/json",
        )

        self.worker_01.refresh_from_db()

        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(self.worker_01.dynamic_metadata, metadata)
        assert self.worker_01.dynamic_metadata_updated_at is not None
        self.assertLessEqual(
            self.worker_01.dynamic_metadata_updated_at, timezone.now()
        )

        self.assertEqual(self.worker_02.dynamic_metadata, {})
        self.assertIsNone(self.worker_02.dynamic_metadata_updated_at)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Gunicorn settings for the Debusine server."""

import os

from gunicorn.arbiter import Arbiter
from gunicorn.workers.base import Worker

from debusine.project.settings.pkg_paths import DEBUSINE_DATA_PATH
from debusine.server.prometheus_client.redis import mark_process_dead

bind = f"unix:{DEBUSINE_DATA_PATH}/debusine.sock"
# Recommended by https://gunicorn.org/design/#scaling, but may need tuning
# for individual systems.
workers = os.cpu_count() * 2 + 1
worker_class = "uvicorn.workers.UvicornWorker"
# Recycle workers every so often to control memory leaks.
max_requests = 500


def child_exit(server: Arbiter, worker: Worker) -> None:  # noqa: ARG001, U100
    """Tell the Prometheus client that a worker has died."""
    mark_process_dead(worker.pid)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""UI helpers for worker pools."""

from functools import cached_property

from debusine.db.models import WorkerPool
from debusine.web.views.places import ResourcePlace
from debusine.web.views.ui.base import UI


class WorkerPoolUI(UI[WorkerPool]):
    """UI helpers for worker pool instances."""

    @cached_property
    def place(self) -> ResourcePlace:
        """Return a place to show this worker pool."""
        return ResourcePlace(
            title=f"Worker pool {self.instance.name}",
            breadcrumb=self.instance.name,
            url=self.instance.get_absolute_url(),
        )
